
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftORleftANDAND GENE LPAREN OR RPAREN UNKNOWNexpression : expression OR expressionexpression : expression AND expressionexpression : LPAREN expression RPARENexpression : GENEexpression : UNKNOWN'
    
_lr_action_items = {'LPAREN':([0,2,5,6,],[2,2,2,2,]),'GENE':([0,2,5,6,],[3,3,3,3,]),'UNKNOWN':([0,2,5,6,],[4,4,4,4,]),'$end':([1,3,4,8,9,10,],[0,-4,-5,-1,-2,-3,]),'OR':([1,3,4,7,8,9,10,],[5,-4,-5,5,-1,-2,-3,]),'AND':([1,3,4,7,8,9,10,],[6,-4,-5,6,6,-2,-3,]),'RPAREN':([3,4,7,8,9,10,],[-4,-5,10,-1,-2,-3,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,5,6,],[1,7,8,9,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> expression OR expression','expression',3,'p_expression_or','gen_parser.py',40),
  ('expression -> expression AND expression','expression',3,'p_expression_and','gen_parser.py',44),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','gen_parser.py',48),
  ('expression -> GENE','expression',1,'p_expression_gene','gen_parser.py',52),
  ('expression -> UNKNOWN','expression',1,'p_expression_unknown','gen_parser.py',56),
]
